/* $Id: rled.c,v 1.2 1993/10/25 23:38:35 cwikla Exp $ */
/*
**
** Copyright (c) 1993 by Wolfram Research, Inc.  All Rights Reserved.
**
** WOLFRAM RESEARCH, INC. ("WRI") DOES NOT WARRANT THAT THE SOFTWARE CONTAINED IN
** THIS FILE IS FREE FROM BUGS AND OMISSIONS; THE SOFTWARE IS PROVIDED AS IS.
** WRI MAKES NO REPRESENTATIONS, EXPRESS OR IMPLIED, WITH RESPECT TO THE SOFTWARE,
** INCLUDING WITHOUT LIMITATIONS ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR
** FITNESS FOR A PARTICULAR PURPOSE, ALL OF WHICH IS EXPRESSLY DISCLAIMED.  WRI
** DOES NOT WARRANT THAT THE SOFTWARE WILL MEET YOUR REQUIREMENTS OR THAT ITS
** OPERATION WILL BE UNINTERRUPTED OR ERROR FREE.
**
** WRI DOES NOT WARRANT THAT THE SOFTWARE CONTAINED IN THIS FILE DOES NOT
** INFRINGE UPON THE INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTIES.
**
** USE OF THE SOFTWARE CONTAINED IN THIS FILE IS ENTIRELY AT YOUR OWN RISK.
**
** To the extent that it is authorized to do so, Wolfram Research grants the
** recipient of this software permission to use the software for his/her own
** personal purposes subject the user's acceptance of the terms stated above
** and the user's inclusion of Wolfram Reseach's copyright notice and disclaimers
** in any reproduction and/or derivative of this file.
**
**
**
**  John L. Cwikla
**  cwikla@wri.com
**  Wolfram Research, Inc
*/

/*
** Simple run length encode/decode stuff. JLC.
*/

#include <X11/Xmd.h>
#include <stdio.h>
#include "rled.h"


#define RBYTE 0x80
#define UNBYTE (0xFF - RBYTE)

#ifndef NULL
#define NULL 0
#endif /* NULL */

#ifndef FALSE
#define FALSE 0
#endif /* FALSE */

#ifndef TRUE
#define TRUE 1
#endif /* TRUE */

unsigned char *RunLengthEncode(unsigned char *_data, INT32 _length, INT32 *_retLength)
{
	unsigned char run;
	unsigned char current, repeat;
	unsigned char *retData, *tmp, *ptr, *dptr; 
	unsigned char haveRepeat;
	int i;
	int swaptest = 1;
	int forgetit;
#if (DEBUGGING > 0)
	double compression;
#endif

	forgetit = FALSE;

	retData = (unsigned char *)calloc(sizeof(unsigned char), _length+sizeof(INT32)+10*sizeof(unsigned char));

	if (retData == (unsigned char *)NULL)
		return retData;

	ptr = retData + sizeof(unsigned char);

	bcopy(&_length, ptr, sizeof(INT32));

	if (*(char *) &swaptest)
		SwapLong((char *)ptr, sizeof(INT32));

	ptr += sizeof(INT32);
	dptr = _data;

	haveRepeat = FALSE;
	run = 1;
	repeat = 0;
	forgetit = FALSE;

	while(!forgetit && (dptr-_data) < _length)
	{
		current = *dptr;
		if ((current == repeat) && (dptr != _data)) /* Do we have a run */
		{
			haveRepeat = TRUE;
			run++;
			if (run == UNBYTE)	/* Are we maxed? */
			{
				run--;
				ptr--;
				*ptr++ = (run | RBYTE);
				*ptr++ = repeat;
				haveRepeat = FALSE;
				run = 1;
				*ptr++ = repeat;

				if ((ptr - (retData + sizeof(INT32) + 1)) > _length)
					forgetit = TRUE;
			}
		}
		else
		{
			if (haveRepeat)
			{
				ptr--;
				*ptr++ = (run | RBYTE);
				*ptr++ = repeat;
				run = 1;
				haveRepeat = FALSE;
			}

			if (current & RBYTE) /* Does it look like a length byte? */ 
			{
				haveRepeat = TRUE;
				run = 1;
			}
			*ptr++ = current;
			repeat = current;
		
			if ((ptr - (retData + sizeof(INT32) + 1)) > _length)
				forgetit = TRUE;
		}
		dptr++;
	}

	if (forgetit)
	{
#if (DEBUGGING > 0)
printf("%d overrun.\n", ptr - (retData + sizeof(INT32) + 1));
#endif
 
		free(retData);
		retData = (unsigned char *)malloc(sizeof(unsigned char) * (_length + sizeof(unsigned char)));
		bcopy(_data, retData+1, _length);
		*_retLength = _length + sizeof(unsigned char);
		*retData = 0;
	}
	else
	{
		if (haveRepeat)
		{
			ptr--;
			*ptr++ = (run | RBYTE);
			*ptr++ = repeat;
		}
		*_retLength = ptr-retData;
		*retData = 0xFF;
	}

#if (DEBUGGING > 0)
	if (*_retLength != _length)
	{
		compression = ((double)*_retLength / (double)_length); 
		printf("Compressed to %d bytes from %d (%.2f).\n", *_retLength, _length, 100.0-compression*100.0);
	}
	else
		printf("No compression.\n");

	fflush(stdout);
#endif /* DEBUGGING */

	return retData;
}

unsigned char *RunLengthDecode(unsigned char *_data, INT32 _length, INT32 *_retLength)
{
	unsigned char run;
	unsigned char current, repeat;
	unsigned char *retData, *ptr, *dptr;
	unsigned char haveRepeat;
	int i, j;
	int retlength;
	int swaptest = 1;
	unsigned char compressed;

	if (!_data)
		return (unsigned char *)NULL;

	compressed = *_data;
/*
** See if data was really compressed.
*/
	if (compressed == 0)
	{
		retData = (unsigned char *)malloc(sizeof(unsigned char *)*(_length - sizeof(unsigned char)));
		*_retLength = _length - sizeof(unsigned char);
		bcopy((_data+sizeof(unsigned char)), retData, *_retLength);
#if (DEBUGGING > 0)
		printf("Data didn't need to be decompressed (%d).\n", *_retLength);
#endif
		return retData;
	}

	dptr = _data + sizeof(unsigned char);

	bcopy(dptr, _retLength, sizeof(INT32));

	if (*(char *) &swaptest)
		SwapLong((char *)_retLength, sizeof(INT32));

	retData = (unsigned char *)calloc(sizeof(unsigned char), *_retLength);

	if (retData == (unsigned char *)NULL)
		return (unsigned char *)NULL;

	ptr = retData; 
	dptr += sizeof(INT32);

	while(dptr-_data < _length) /* ptr-retData < length) */
	{
		if (*dptr >= RBYTE)
		{
			run = (*dptr & UNBYTE);
			dptr++;
			for(i=0;i<run;i++)
				*ptr++ = *dptr;
		}
		else
			*ptr++ = *dptr;

		dptr++;
	}

#if (DEBUGGING > 0)
	if (*_retLength != (dptr - _data))
		printf("Decompressed from %d to %d.\n", dptr - _data, *_retLength);
	else
		printf("No decompression.\n");

	fflush(stdout);
#endif /* DEBUGGING */

	return retData;
}
